;; mysql.lsp - newLISP connection to the MySql database
;;
;; version 2.2  (c) Lutz Mueller 2003, licensed GPL v2
;;              (c) Gordon Fischer 2005
;;
;; current tested MySQL version is 4.1.14
;;
;; version 2.0 changed offsets for MySQL version 4.0
;;     functions now check for valid result set to avoid crashing
;;     eliminated 'eof' which did not work (always return true)
;;
;; version 2.1 'inserted-id' was misspelled, revised test routine
;; version 2.1 'get-integer' yo 'get-int'
;;
;; version 2.2 'inserted-id' offset changed to libmysqlclient call
;; version 2.2 'keep_type' added to map MySQL types to newLISP types
;; version 2.2 'fetch-row' bug when column contained NULL
;;             added code to return nil when MySQL column contains NULL
;; version 2.2 'query' enhancement - basic identification of insert statement
;;             query now returns inserted row id
;;
;; version 2.3 now safe for newLISP with reconfigured ops: +,-,*,/ to
;;             add,sub,mul,div
;;
;; 'close' in MAIN
;;
;; See also http://www.mysql.com
;; Import functions from MySql library 'libmysqlclient.so'
;; This library might be in a different location on a particular
;; installation of MySql or have a different name
;; (change accordingly below)
;;
;; This file is tested with MySQL 4.0/4.1  and will not work on
;; previous versions 3.22 and 3.23 which need different offsets
;; into the MYSQL result data structure.
;;
;; Compile and run 'sql.c' to get the correct offsets for
;; previous versions of MySQL. (for num-rows, error, affected-rows)
;;
;; Some of the functions like myswl_num_rows() cannot be imported
;; because they are really macros extracting data from structures
;; like MYSQL or MYSQL_RES (see mysql.h) in your distribution)
;;
;; The file sql.c in the newLISP distribution contains a programs
;; calculating the offsets of most important fields in these
;; structures. Those offsets are used here to retrieve values for
;; the 'number of rows' in a result set, etc. . Using these offsets
;; and the information found in 'mysql.h' and 'mysql_com.h', other
;; functions can be imported and wrappers built around them.
;; In this case you need to install the developer's version of
;; MySQL to get the header filers mentioned.
;; On a delivered system only the shared libraries specifically
;; 'libmysqlclient.so' are needed.
;;
;;
;; see 'test-mysql' at the bottom of this page for a short demo
;; how to use the functions.
;; 
;;     MySQL:init ................ get a database handle MYSQL
;;     MySQL:connect ............. connect to a database
;;     MySQL:query ............... execute a SQL statement
;;     MySQL:num-rows ............ rows in result of query
;;     MySQL:num-fields .......... columns in result of query
;;     MySQL:fetch-row ........... get row from the query result
;;     MySQL:fetch-all ........... get all rows from the last query
;;     MySQL:database ............ return all database names
;;     MySQL:tables .............. return all tables names
;;     MySQL:fields .............. return all fields in a table
;;     MySQL:data-seek ........... position in result for fetching
;;     MySQL:affected-rows ....... number of affected rows from operation
;;     MySQL:inserted-id ......... last value of auto increment id operation
;;     MySQL:error ............... get error message
;;     MySQL:close-db ............ close database connection
;;

(context 'MySQL)

(set 'libmysqlclient "/usr/lib/libmysqlclient.so.14")

(import libmysqlclient "mysql_init")
(import libmysqlclient "mysql_real_connect")
(import libmysqlclient "mysql_get_host_info")
(import libmysqlclient "mysql_escape_string")
(import libmysqlclient "mysql_query")
(import libmysqlclient "mysql_real_query")
(import libmysqlclient "mysql_store_result")
(import libmysqlclient "mysql_free_result")
(import libmysqlclient "mysql_data_seek")
(import libmysqlclient "mysql_fetch_row")
(import libmysqlclient "mysql_close")
(import libmysqlclient "mysql_fetch_field_direct")
(import libmysqlclient "mysql_insert_id")

;; the followong definitions wrap the imported mysql functions to yield
;; more usable and higher level lisp functions. All the MySQL:xxx functions
;; defined return 'true' on success or 'nil' on failure.

;; (init)                             - init database access and access handle
;; (connect host user passw database) - connect to a database and authenticate
;; (query sql)                        - make an SQL query or update
;; (num-rows)                         - get the number of rows of last query
;; (num-fields)                       - get number of fields/columns in query
;; (fetch-row)                        - fetch a row from the query result table
;; (fetch-all)                        - fetch all rows at once
;; (databases)                        - get names of all the existing databases
;; (tables)                           - get a list of tables in active database
;; (fields table)                     - get a list of all fields in a table
;; (data-seek offset)                 - position to a specific row before fetch
;; (error)                            - get error information
;; (affected-rows)                    - no of affected rows in last query/update
;; (inserted-id)                      - last id from auto-increment field
;; (close-db)                         - close a database

;; initialize database access, get an access handle MYSQL
;;
(define (init)
  (set 'MYSQL (mysql_init 0))
  (if (= MYSQL 0) (set 'MYSQL nil))
  (not (= MYSQL nil)))

;; connect to a specific database
;;
(define (connect host user passw database)
  (not(= (mysql_real_connect MYSQL host user passw database 0 0 0) 0)))

;; send a SQL string for evaluation
;;
(define (query sql)
  (if MYSQL_RES (mysql_free_result MYSQL_RES))
  (set 'result  (= (mysql_query MYSQL sql) 0))
  (set 'MYSQL_RES (mysql_store_result MYSQL)) 
  (if (= MYSQL_RES 0) (set 'MYSQL_RES nil))
  (if (and result (find "insert into" sql 1)) (set 'result (inserted-id)))
  result)

;; number of rows from last query
;;
(define (num-rows)
  (if MYSQL_RES (get-int MYSQL_RES)))

;; number of columns from last query
;;
(define (num-fields)
  (if MYSQL_RES (get-int (int (+ MYSQL_RES 60)))))


;; format the result based on the field type.
;;

(define (keep-type res_ptr field_addr column_num, data)
  (set 'type_ptr (mysql_fetch_field_direct res_ptr (int column_num)))
  ;; The field type is the 20th field of the MySQL_FIELD structure
  ;; since fields 1-19 are all 4 byte fields we get the enum value
  ;; like so
  (set 'data (get-int (int (+ type_ptr (* 19 4)))))
  ;; Consult 'enum_field_types' in mysql_com.h for values
  (if (= data 1) ;; boolean
        (get-string field_addr)
      (= data 3) ;; integer
        (int (get-string field_addr))
      (= data 12) ;; datetime
        (apply date-value (map int (parse (get-string field_addr) "[-: ]" 0)))
      (= data 4) ;; float
        (float (get-string field_addr))
      ;; else
      (get-string field_addr)
  )
)

;; fetch a row from a query, subsequent calls go trough the table
;; until the bottom is reached
;;
(define (fetch-row)
  (if MYSQL_RES
    (set 'rdata (mysql_fetch_row MYSQL_RES))
    (set 'rdata 0))
  (if (!= rdata 0)
    (begin
      (set 'row '())
      (dotimes (field (num-fields))
            (set 'field_addr (get-int (int (+ rdata (* field 4)))))
            (if (= field_addr 0)
              (push nil row -1) ;; what to do when the field contains NULL
              (push (keep-type MYSQL_RES field_addr field) row -1)))
    row)))

;; return all rows/columns from the last query
;;
(define (fetch-all)
  (set 'all '())
  (dotimes (x (num-rows)) (push (fetch-row) all))
  (reverse all))

;; get a list of databases
;;
(define (databases)
  (query "show databases;")
  (fetch-all))

;; get a list of tables
;;
(define (tables)
  (query "show tables;")
  (fetch-all))

;; get a list of fields
;;
(define (fields table)
  (query (append "show fields from " table ";"))
  (fetch-all))
  

;; position to row in result set
;;
(define (data-seek offset)
  (if MYSQL_RES
    (begin
    (mysql_data_seek MYSQL_RES offset)
    true)))

;; text info about error
;;
(define (error)
  (if MYSQL (get-string (+ MYSQL 60))))

;; number of affected rows by last MySQL:query operation
;;
(define (affected-rows)
  (if MYSQL
    (get-int (int (+ MYSQL 368)))))

;; last insert_id from auto increment field
;;
(define (inserted-id)
  (if MYSQL (mysql_insert_id MYSQL)))

;; close database access
;;
(define (close-db)
  (if MYSQL_RES (mysql_free_result MYSQL_RES))
  (if MYSQL (mysql_close MYSQL))
  (set 'MYSQL nil)
  (set 'MYSQL_RES nil)
  true)

(context 'MAIN)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; test data base functions
;;

(define (test-mysql)
  (MySQL:init)
  (MySQL:connect 0 "" 0 "test")
  
  (println "databases:")
  (MySQL:query "show databases;")
  (dotimes (x (MySQL:num-rows)) (println (MySQL:fetch-row)))
  (println)
  
  (MySQL:query "create table fruits (name CHAR(20), 
                                     qty INT(3),
                                     num INT(4) AUTO_INCREMENT UNIQUE);")
  
  (MySQL:query "insert into fruits values ('apples', 11, null);")
  (println "inserted-id: " (MySQL:inserted-id))
  (MySQL:query "insert into fruits values ('oranges', 22, null);")
  (println "inserted-id: " (MySQL:inserted-id))
  (MySQL:query "insert into fruits values ('bananas', 33, null);")
  (println "inserted-id: " (MySQL:inserted-id))
  
  (println "\n" (MySQL:affected-rows) " affected rows in query")

  (println "inserted into fruits:")
  (MySQL:query "select * from fruits;")
  (dotimes (x (MySQL:num-rows)) (println (MySQL:fetch-row)))

  (println "rows = " (MySQL:num-rows) " fields = " (MySQL:num-fields))
  (println)
  
  (println "tables:")
  (MySQL:query "show tables;")
  (dotimes (x (MySQL:num-rows)) (println (MySQL:fetch-row)))
  (println)
  
  (MySQL:query "select * from fruits;")
  (MySQL:data-seek 2)
  (println "data-seek to offset 2:")
  (println (MySQL:fetch-row))
  (println)
  
  (println "fetch-all:")
  (println (MySQL:query "select * from fruits;"))
  (println (MySQL:fetch-all))   
  
  (MySQL:query "drop table fruits;")
  (MySQL:close-db)
)

;; eof
