;; mysql5.lsp - newLISP connection to the MySQL 5  database
;;
;; version 2.41  (c) Lutz Mueller 2003-2006, licensed GPL v2
;;               (c) Gordon Fischer 2005
;;
;; Configured and tested MySQL version is 5.0.19, safe for newLISP
;; operators +,-,*,/ configured as add,sub,mul,div
;;
;; Note that this implementation supports a maximum of 2,147,483,647
;; rows in a database table. Now automatically adjusts row indexes to
;; endian type of host CPU, but higher 32 bits are treated as 0 for now.
;;
;; See also http://www.mysql.com
;; Import functions from MySql library:
;; '/usr/local/mysql/libmysqlclient.15.so' on LINUX/UNIX
;; '/usr/local/mysql/libmysqlclient.15.dylib on MacOS X
;;
;; to compile MySQL with client libraries use:
;; ./configure --prefix=/usr/local --enable-shared
;;`
;; This library might be in a different location on a particular
;; installation of MySql or have a different name
;; (change accordingly below)
;;
;; Compile and run 'sql.c' to get the correct offsets for
;; previous versions of MySQL. (for num-rows, error, affected-rows)
;;
;; Some of the functions like myswl_num_rows() cannot be imported
;; because they are really macros extracting data from structures
;; like MYSQL or MYSQL_RES (see mysql.h) in your distribution)
;;
;; The file sql.c in the newLISP distribution contains a programs
;; calculating the offsets of most important fields in these
;; structures. Those offsets are used here to retrieve values for
;; the 'number of rows' in a result set, etc. . Using these offsets
;; and the information found in 'mysql.h' and 'mysql_com.h', other
;; functions can be imported and wrappers built around them.
;; In this case you need to install the developer's version of
;; MySQL to get the header filers mentioned.
;; On a delivered system only the shared libraries specifically
;; 'libmysqlclient.so' are needed.
;;
;;
;; see 'test-mysql' at the bottom of this page for a short demo
;; how to use the functions.
;; 
;;     MySQL:init ................ get a database handle MYSQL
;;     MySQL:connect ............. connect to a database
;;     MySQL:query ............... execute a SQL statement
;;     MySQL:num-rows ............ rows in result of query
;;     MySQL:num-fields .......... columns in result of query
;;     MySQL:fetch-row ........... get row from the query result
;;     MySQL:fetch-all ........... get all rows from the last query
;;     MySQL:database ............ return all database names
;;     MySQL:tables .............. return all tables names
;;     MySQL:fields .............. return all fields in a table
;;     MySQL:data-seek ........... position in result for fetching
;;     MySQL:affected-rows ....... number of affected rows from operation
;;     MySQL:inserted-id ......... last value of auto increment id operation
;;     MySQL:error ............... get error message
;;     MySQL:close-db ............ close database connection
;;

(context 'MySQL)

;(set 'libmysqlclient "/usr/lib/libmysqlclient.so")
(set 'libmysqlclient "/usr/local/mysql/lib/libmysqlclient.dylib")

(import libmysqlclient "mysql_init")
(import libmysqlclient "mysql_real_connect")
(import libmysqlclient "mysql_get_host_info")
(import libmysqlclient "mysql_escape_string")
(import libmysqlclient "mysql_query")
(import libmysqlclient "mysql_real_query")
(import libmysqlclient "mysql_store_result")
(import libmysqlclient "mysql_free_result")
(import libmysqlclient "mysql_data_seek")
(import libmysqlclient "mysql_fetch_row")
(import libmysqlclient "mysql_close")
(import libmysqlclient "mysql_fetch_field_direct")
(import libmysqlclient "mysql_insert_id")

;; following offsets into 'C' data structures are different on each mayor MySQL
;; version compile and run util/sql.c from the distribution to obtain these numbers

;; check endianess of the host CPU 
(set 'big-endian (= (pack ">ld" 1) (pack "ld" 1)))

(constant 'NUM_ROWS_OFFSET (if big-endian 4 0))
(constant 'NUM_FIELDS_OFFSET 60)
(constant 'ERROR_OFFSET 85)
(constant 'INSERT_ID_OFFSET (if big-endian 708 704))
(constant 'AFFECTED_ROWS_OFFSET (if big-endian 700 696))

;; the followong definitions wrap the imported mysql functions to yield
;; more usable and higher level lisp functions. All the MySQL:xxx functions
;; defined return 'true' on success or 'nil' on failure.

;; (init)                             - init database access and access handle
;; (connect host user passw database) - connect to a database and authenticate
;; (query sql)                        - make an SQL query or update
;; (num-rows)                         - get the number of rows of last query
;; (num-fields)                       - get number of fields/columns in query
;; (fetch-row)                        - fetch a row from the query result table
;; (fetch-all)                        - fetch all rows at once
;; (databases)                        - get names of all the existing databases
;; (tables)                           - get a list of tables in active database
;; (fields table)                     - get a list of all fields in a table
;; (data-seek offset)                 - position to a specific row before fetch
;; (error)                            - get error information
;; (affected-rows)                    - no of affected rows in last query/update
;; (close-db)                         - close a database

;; initialize database access, get an access handle MYSQL
;;
(define (init)
  (set 'MYSQL (mysql_init 0))
  (if (= MYSQL 0) (set 'MYSQL nil))
  (not (= MYSQL nil)))

;; connect to a specific database
;;
(define (connect host user passw database)
  (not(= (mysql_real_connect MYSQL host user passw database 0 0 0) 0)))

;; send a SQL string for evaluation
;;
(define (query sql)
  (if MYSQL_RES (mysql_free_result MYSQL_RES))
  (set 'result  (= (mysql_query MYSQL sql) 0))
  (set 'MYSQL_RES (mysql_store_result MYSQL)) 
  (if (= MYSQL_RES 0) (set 'MYSQL_RES nil))
  (if (and result (find "insert into" sql 1)) (set 'result (inserted-id)))
  result)

;; number of rows from last query
;;
(define (num-rows)
  (if MYSQL_RES (get-int (int (+ MYSQL_RES NUM_ROWS_OFFSET)))))

;; number of columns from last query
;;
(define (num-fields)
  (if MYSQL_RES (get-int (int (+ MYSQL_RES NUM_FIELDS_OFFSET)))))


;; format the result based on the field type.
;;

(define (keep-type res_ptr field_addr column_num, data)
  (set 'type_ptr (mysql_fetch_field_direct res_ptr (int column_num)))
  ;; The field type is the 20th field of the MySQL_FIELD structure
  ;; since fields 1-19 are all 4 byte fields we get the enum value
  ;; like so
  (set 'data (get-int (int (+ type_ptr (* 19 4)))))
  ;; Consult 'enum_field_types' in mysql_com.h for values
  (if (= data 1) ;; boolean
        (get-string field_addr)
      (= data 3) ;; integer
        (int (get-string field_addr))
      (= data 12) ;; datetime
        (apply date-value (map int (parse (get-string field_addr) "[-: ]" 0)))
      (= data 4) ;; float
        (float (get-string field_addr))
      ;; else (will handle TEXT type 252)
      (get-string field_addr)
  )
)

;; fetch a row from a query, subsequent calls go trough the table
;; until the bottom is reached
;;
(define (fetch-row)
  (if MYSQL_RES
    (set 'rdata (mysql_fetch_row MYSQL_RES))
    (set 'rdata 0))
  (if (!= rdata 0)
    (begin
      (set 'row '())
      (dotimes (field (num-fields))
            (set 'field_addr (get-int (int (+ rdata (* field 4)))))
            (if (= field_addr 0)
              (push nil row -1) ;; what to do when the field contains NULL
              (push (keep-type MYSQL_RES field_addr field) row -1)))
    row)))

;; return all rows/columns from the last query
;;
(define (fetch-all)

  (dotimes (x (num-rows)) (push (fetch-row) all))
  (reverse all))

;; get a list of databases
;;
(define (databases)
  (query "show databases;")
  (fetch-all))

;; get a list of tables
;;
(define (tables)
  (query "show tables;")
  (fetch-all))

;; get a list of fields
;;
(define (fields table)
  (query (append "show fields from " table ";"))
  (fetch-all))
  

;; position to row in result set
;;
(define (data-seek offset)
  (if MYSQL_RES
    (if big-endian
    	(mysql_data_seek MYSQL_RES  0 offset)
        (mysql_data_seek MYSQL_RES offset 0))))

;; text info about error
;;
(define (error)
  (if MYSQL (get-string (+ MYSQL ERROR_OFFSET))))

;; number of affected rows by last MySQL:query operation
;;
(define (affected-rows)
  (if MYSQL
    (get-int (int (+ MYSQL AFFECTED_ROWS_OFFSET)))))

;; last insert_id from auto increment field
;;
(define (inserted-id)
;  (if MYSQL (mysql_insert_id MYSQL))
  (if MYSQL (get-int (int (+ MYSQL INSERT_ID_OFFSET)))))

;; close database access
;;
(define (close-db)
  (if MYSQL_RES (mysql_free_result MYSQL_RES))
  (if MYSQL (mysql_close MYSQL))
  (set 'MYSQL nil)
  (set 'MYSQL_RES nil)
  true)

(context 'MAIN)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; test data base functions
;;

(define (test-mysql)
  (MySQL:init)
  (MySQL:connect 0 "" 0 "test")
  
  (println "databases:")
  (MySQL:query "show databases;")

  (dotimes (x (MySQL:num-rows)) (println (MySQL:fetch-row)))
  (println)
  
  (MySQL:query "create table fruits (name TEXT(2000), 
                                     qty INT(3),
                                     num INT(4) AUTO_INCREMENT UNIQUE);")
  
  (MySQL:query "insert into fruits values ('apples', 11, null);")
  (println "inserted-id: " (MySQL:inserted-id))
  (MySQL:query "insert into fruits values ('oranges', 22, null);")
  (println "inserted-id: " (MySQL:inserted-id))
  (MySQL:query "insert into fruits values ('bananas', 33, null);")
  (println "inserted-id: " (MySQL:inserted-id))

  (println "inserted into fruits:")
  (MySQL:query "select * from fruits;")
  (println "\n" (MySQL:affected-rows) " affected rows in query select")
  (dotimes (x (MySQL:num-rows)) (println (MySQL:fetch-row)))

  (println "rows = " (MySQL:num-rows) " fields = " (MySQL:num-fields))
  (println)
  
  (println "tables:")
  (MySQL:query "show tables;")
  (dotimes (x (MySQL:num-rows)) (println (MySQL:fetch-row)))
  (println)
  
  (MySQL:query "select * from fruits;")
  (MySQL:data-seek 2)

  (println "data-seek to offset 2:")
  (println (MySQL:fetch-row))
  (println)
  
  (println "fetch-all:")
  (println (MySQL:query "select * from fruits;"))
  (println (MySQL:fetch-all))   
  
  (MySQL:query "drop table fruits;")
  (MySQL:close-db)
)

;; correct output from (test-mysql)

;    newLISP v.8.8.0 on OSX UTF-8, execute 'newlisp -h' for more info.
;
;    > (test-mysql)
;    databases:
;    ("information_schema")
;    ("test")

;    inserted-id: 1
;    inserted-id: 2
;    inserted-id: 3
;    inserted into fruits:

;    3 affected rows in query select
;    ("apples" 11 1)
;    ("oranges" 22 2)
;    ("bananas" 33 3)
;    rows = 3 fields = 3

;    tables:
;    ("fruits")

;    data-seek to offset 2:
;    ("bananas" 33 3)

;    fetch-all:
;    true
;    (("apples" 11 1) ("oranges" 22 2) ("bananas" 33 3))
;    true
;    > 

;; eof
