;; odbc.lsp - OCBC Interface functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 
;; this has only been tested on Win32  but should work on UNIX too
;; 
;; some of the code assumes Intel (low -> high) little-endian byte order
;;
;; version 1.1 July 30th, 2003
;; 2 new functions ODBC:tables and  ODBC:columns contributed by Steve Adams
;;
;; version 1.2 August 28th, 2003
;; fetch didn't initialise row variables, so old value would be copied if
;; afield was empty in a fetched row
;;
;; version 1.3 September 22nd, 2003
;; one additional space allocated in var initialization
;;
;; version 1.4 2004-01-08
;; changed import, failure is handled by newLISP error message
;; now trims result fields in rows in 'test-odbc'
;;
;; version 1.5 'get-integer' to 'get-int'
;; version 1.6 'integer' to 'int'
;;
;; See the end of file for documentation with each user function definition
;; and a test function, which demos the functionality
;; 
;;  (ODBC:connect data-source-name-str user-name-str password-str) ; connect to a data source
;; 
;;  (ODBC:query sql-str)          ; perform a SQL statement
;;
;;  (ODBC:num-cols)               ; number of columns in a sql query result set with 'select'
;;
;;  (ODBC:column-atts col)        ; retrieve columns attributes
;;
;;  (ODBC:fetch-row)              ; fetch a row of data after a sql query with 'select'
;;
;;  (ODBC:affected-rows)          ; # rows affected by a sql query: 'delete', 'update' etc.
;;
;;  (ODBC:close-db)               ; close database connection
;;
;;  (ODBC:tables)                 ; returns a list of tables in the current database
;;
;;  (ODBC:columns table-name)     ; returns an array of column attributes in table-name
;;                                  one row for each columns with name, datatype and precision 
;;

(context 'ODBC)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; start up importing functions ;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;; set to the appropiate library on Unix or Win32

(define ODBC-library "odbc32.dll")

;; Constants used, make sure these constants are Ok on your Operating System or Platform.
;; Note, that (define var value) is the same as as saying (set 'var value), it is here more
;; of a visual distinction, documenting that values are constants and shouldn't be changed.
;; Most of these are defned in sql.h, sqltypes.h and sqlext.h of your platform.
;; The following definitions come from c:\Borland\BCC\Include

(define SQL_HANDLE_ENV          1)
(define SQL_HANDLE_DBC          2)
(define SQL_HANDLE_STMT         3)
(define SQL_HANDLE_DESC         4)

(define SQL_NULL_HANDLE         0)

(define SQL_SUCCESS             0)
(define SQL_SUCCESS_WITH_INFO   1)

(define SQL_OV_ODBC3            3)
(define SQL_ATTR_ODBC_VERSION	200)

(define SQL_LOGIN_TIMEOUT     103)

(define SQL_NTS                -3)

(define SQL_CHAR                1)
(define SQL_C_CHAR       SQL_CHAR)


;; Import functions
;; there are many more, which are not used here, goto microsoft.com and unixodbc.org for
;; more information on ODBC SQLxxx API


(set 'funcs '(
	"SQLAllocHandle"
	"SQLSetEnvAttr"
	"SQLFreeHandle"
	"SQLSetConnectAttr"
	"SQLConnect"
	"SQLDisconnect"
	"SQLGetDiagRec"
	"SQLExecDirect"
	"SQLNumResultCols"
	"SQLRowCount"
	"SQLBindCol"
	"SQLFetch"
	"SQLDescribeCol"
	"SQLTables"
	"SQLColumns"))

(dolist (fun funcs)
	(import ODBC-library fun))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; reserve space for global pointers ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(set 'ptr-odbc-env "    ")     ; pointer to environment handle
(set 'ptr-odbc-conn "    ")    ; pointer to connection handle
(set 'ptr-result-cols "    ")  ; pointer to number of columns in result
(set 'ptr-odbc-stmt "    ")    ; pointer to handle for sql statement
(set 'ptr-result-rows "    ")  ; pointer to number of affected rows from sql statement

(set 'odbc-stmt nil)           ; statement handle
(set 'odbc-conn nil)           ; connection handle
(set 'result-cols 0)           ; contains the number of rows resulting from a 'select' qery

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; AUXILIARY ROUTINES ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; check result code ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (is-error-result)
	;;result is 16bit, disregard upper 16 bits
	(set 'odbc-result (& 0xFFFF odbc-result))
	(and (!= odbc-result SQL_SUCCESS) (!= odbc-result SQL_SUCCESS_WITH_INFO)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; initialize and make connection ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (init)
	(and
		;; get environment handle
		(set 'odbc-result (SQLAllocHandle SQL_HANDLE_ENV SQL_NULL_HANDLE ptr-odbc-env))

		(if (is-error-result)
			(begin
				(println "Error allocating env handle")
				nil) true)

		(set 'odbc-env (get-int ptr-odbc-env))

		;; register version
		(set 'odbc-result (SQLSetEnvAttr odbc-env SQL_ATTR_ODBC_VERSION SQL_OV_ODBC3 0))

		(if (is-error-result)
			(begin
				(println "Error setting ODBC environment")
				(SQLFreeHandle SQL_HANDLE_ENV odbc-env)
				nil) true))
	)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; get diagnostic record ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; retrieve error info after last failed ODBC request
;;
;; type is one of the following:
;;
;; SQL_HANDLE_ENV, SQL_HANDLE_DBC, SQL_HANDLE_STMT, SQL_HANDLE_DESC 
;;

(define (error type)
	(set 'diag-status "          ")
	(set 'diag-err  "    ")
	(set 'diag-mlen "    ")
	(set 'diag-message "                                                                ")
	(SQLGetDiagRec type odbc-conn 1 diag-status diag-err diag-message 64 diag-mlen)
	(string diag-message " " diag-status (get-int diag-err)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; bind all columns to string output ;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; before fetching rows string variables are configured with sufficient long string buffers
;; for the 'fetch' statement.
;;

(set 'vars '(var0  var1  var2  var3  var4  var5  var6  var7  var8  var9 
             var10 var11 var12 var13 var14 var15 var16 var17 var18 var19 
             var20 var21 var22 var23 var24 var25 var26 var27 var28 var29
             var30 var32 var32 var33 var34 var35 var36 var37 var38 var39
             var40 var41 var42 var43 var44 var45 var46 var47 var48 var49
             var50 var51 var52 var53 var54 var55 var56 var57 var58 var59
             var60 var51 var62 var63 var64))


(define (bind-columns)
	(set 'ptr-result-err "    ")
	(for (v 1 result-cols)
		(set 'w (+ (last (column-atts v)) 1))
		(set (nth v vars) (format (string "%" w "s") ""))
		(SQLBindCol odbc-stmt (int v) SQL_C_CHAR (eval (nth v vars)) w ptr-result-err))
	
	true)


;====================================  USER ROUTINES ========================================


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; connect to a data source ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;
;; connect to a data source with a user name and password
;; the data-source name must be configured first via ODBC
;; administrative software, i.e. control panel on Win32
;;
;; example: 
;;           (connect "mydatabase" "johndoe" "secret")
;;
;;


(define (ODBC:connect data-source user password)

	(and
		(init)

		;; allocate connection handle
		(set 'odbc-result (SQLAllocHandle SQL_HANDLE_DBC odbc-env ptr-odbc-conn))

		(if (is-error-result)
			(begin
				(println "Error allocating conn handle")
				(SQLFreeHandle SQL_HANDLE_ENV odbc-env)
				nil) true)

		(set 'odbc-conn (get-int ptr-odbc-conn))

		;; set timeout for connection
		(SQLSetConnectAttr odbc-conn SQL_LOGIN_TIMEOUT 5 0)

		;; connect to a data source
		(set 'odbc-result (SQLConnect odbc-conn data-source SQL_NTS
                                                      user SQL_NTS
                                                      password SQL_NTS))

		(if (is-error-result)
			(begin
				(println "Could not connect")
				(SQLFreeHandle SQL_HANDLE_DBC odbc-conn)
				(SQLFreeHandle SQL_HANDLE_ENV odbc-env)
				nil) true))
	)




;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; make a query ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; send and SQL string for database manipulation
;;
;; example:
;;          (query "select * from someTable")
;;          (query "delete from addresses")
;;          (query "insert into fruits values ('apples', 11)")
;;                     - etc -
;;
;;
;;

(define (ODBC:query sql-string)
	(and 
		;; is stmt handle exists free it
		(if odbc-stmt (begin 
			(SQLFreeHandle SQL_HANDLE_STMT odbc-stmt)
			(set 'odbc-stmt nil)
			true)	true)

		;; allocate statement handle
		(set 'odbc-result (SQLAllocHandle SQL_HANDLE_STMT odbc-conn ptr-odbc-stmt))

		(if (is-error-result)
			(begin
				(println "could not allocate statement handle")
				nil)
			(set 'odbc-stmt (get-int ptr-odbc-stmt)))

		;; do the query
		(set 'odbc-result (SQLExecDirect odbc-stmt sql-string SQL_NTS))
		(if (is-error-result)
			(begin
				(println "query failed")
				nil)
			true)

		;; find number of columns in result set
		(SQLNumResultCols odbc-stmt ptr-result-cols)
		(set 'result-cols (& 0xFFFF (get-int ptr-result-cols)))

		;; bind colums to string vars for fetching
		(if (not (= result-cols 0)) (bind-columns) true)
		true
		)
		
	)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Returns the number columns in a 'select' query result set
;;
;; example:
;;          (ODBC:num-cols)
;;

(define (num-cols) result-cols)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; get data type/width in columns ;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; returns a list with the columname SQL data type number and required column size
;; when displaying in a string. For the data type number and SQL data type see
;; the file sql.h on your platform OS, i.e. SQL_VARCHAR, SQL_INTEGER etc.
;;
;; before using column-atts a query has to be performed.
;;
;; example:
;;          (ODBC:column-atts 1) 
;;                => ("name" 12 20)
;;
;; the first column has the header "name" data type SQL_VARCHAR with
;; a maximum of 20 characters.
;;
;;


(define (column-atts col)
	(set 'col-name-out "                                ")
	(set 'ptr-name-len "    ")
	(set 'ptr-data-type "    ")
	(set 'ptr-col-size "    ")
	(set 'ptr-dec-dig "    ")
	(set 'ptr-nullable "    ")

	(set 'odbc-result (& 0xFFFF (SQLDescribeCol odbc-stmt (int col)
                                                col-name-out 32
                                                ptr-name-len
                                                ptr-data-type
                                                ptr-col-size
                                                ptr-dec-dig
                                                ptr-nullable)))
	(list col-name-out (& 0xFFFF (get-int ptr-data-type)) (get-int ptr-col-size)))



;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; fetch a result row ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Fetches a row of data after a previously executed query. Each data is formatted as
;; a string. And can be converted using newLISP conversion functions like:
;;
;; integer, float, string
;;
;; if the data type are unknown then 'column-atts' can be used to retrive the data type
;; number.
;;
;; example:
;;          (ODBC:fetch-row)
;;                 => ("apples" "11")
;;

(define (fetch-row , row)
	(bind-columns)
	(set 'odbc-result (& 0xFFFF (SQLFetch odbc-stmt)))
	(if (is-error-result) 
		nil
		(begin
			(for (x result-cols 1) (push (eval (nth x vars)) row))
			row)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; get number of rows in result set ;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Returns the number of rows affected by a 'insert', 'update' or 'delete' query
;; operation. After a 'select' operation the number -1 will be returned.
;;
;; example:
;;          (ODBC:affected-rows)
;;
;;

(define (affected-rows)	
	(set 'odbc-result (& 0xFFFF (SQLRowCount odbc-stmt ptr-result-rows)))
	(if (is-error-result) 0	(get-int ptr-result-rows)))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; get tables in a database ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Returns a list of tables in the database of the current connection
;;
;; example:
;;          (ODBC:tables)
;;
;;

(define (tables)
    (if (and
        ;; is stmt handle exists free it
        (if odbc-stmt (begin
            (SQLFreeHandle SQL_HANDLE_STMT odbc-stmt)
            (set 'odbc-stmt nil)
            true)   true)

        ;; allocate statement handle
        (set 'odbc-result (SQLAllocHandle SQL_HANDLE_STMT odbc-conn ptr-odbc-stmt))
        (if (is-error-result)
            (begin
                (println "could not allocate statement handle")
                nil)
            (set 'odbc-stmt (get-int ptr-odbc-stmt)))

        ;; do the query
        (set 'odbc-result (SQLTables odbc-stmt 0 SQL_NTS 0 SQL_NTS "%" SQL_NTS 0 SQL_NTS))
        (if (is-error-result)
            (begin
                (println "query failed")
                nil)
            true)

        ;; find number of columns in result set
        (SQLNumResultCols odbc-stmt ptr-result-cols)
        (set 'result-cols (& 0xFFFF (get-int ptr-result-cols)))

        ;; bind colums to string vars for fetching
        (if (not (= result-cols 0)) (bind-columns) true)

        (begin
           (set 'names nil)
           (while (set 'row (ODBC:fetch-row))
               (push (nth 2 row) names -1))
           true)
        ) names)
    )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; get columns of a table ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; 
;; Returns an array of columns and atrributes like nmae, data type and precision, 
;; one row for each colum. The first field in a row is the name 
;;
;; example:
;;         (ODBC:columns "MyTable")
;;
;;

(define (ODBC:columns table)
    (if (and
        ;; is stmt handle exists free it
        (if odbc-stmt (begin
            (SQLFreeHandle SQL_HANDLE_STMT odbc-stmt)
            (set 'odbc-stmt nil)
            true)   true)

        ;; allocate statement handle
        (set 'odbc-result (SQLAllocHandle SQL_HANDLE_STMT odbc-conn ptr-odbc-stmt))

        (if (is-error-result)
            (begin
                (println "could not allocate statement handle")
                nil)
            (set 'odbc-stmt (get-int ptr-odbc-stmt)))

        ;; do the query
        (set 'odbc-result (SQLColumns odbc-stmt 0 SQL_NTS 0 SQL_NTS
                          table SQL_NTS 0 SQL_NTS))
        (if (is-error-result)
            (begin
                (println "query failed")
                nil)
            true)

        ;; find number of columns in result set
        (SQLNumResultCols odbc-stmt ptr-result-cols)
        (set 'result-cols (& 0xFFFF (get-int ptr-result-cols)))

        ;; bind colums to string vars for fetching
        (if (not (= result-cols 0)) (bind-columns) true)

        (begin
           (set 'names nil)
           (while (set 'col (ODBC:fetch-row))
               (set 'attr (list (nth 3 col) (nth 5 col) (nth 6 col) (nth 8 col)))
               (push attr names -1))
           true)
        ) names)
    )


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; close connection ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; closes a database connection
;;
;; example:
;;         (ODBC:close-db)
;;

(define (close-db)
	(if odbc-stmt (SQLFreeHandle SQL_HANDLE_STMT odbc-stmt))
	(set 'odbc-stmt nil)
	(if odbc-conn (begin
		(SQLDisconnect odbc-conn)
		(SQLFreeHandle SQL_HANDLE_DBC odbc-conn)
		(set 'odbc-conn nil)))
	true)
    

(context 'MAIN)
;=================================== test =================================================
;;
;; Note: before performing this test a database with name 'test'
;; and data source name 'test' should be created. The data base
;; should contain a table:
;;
;;      create table fruits (name CHAR(20), qty INT(3))
;;
;; On some systems this can be done using (query "create ....")
;; on MS-Acces this will not work and the table has to be created
;; manually.
;;
;;
;; example:
;;          (test-odbc)
;;



(define (test-odbc)

	;; Note, on MS-Access must create table fruits manually first
	;; else you could do:
	;;   (ODBC:query "create table fruits (name CHAR(20), qty INT(3))")
	;; for "aUser" and "secret" you may just put empty strings ""
	;; i.e. (ODBC:connect "test" "" "")
	;; when on Windows on the same machine

	(if (not (ODBC:connect "test-db" "" "")) (exit))

	(println "connected ...")

	(ODBC:query "insert into fruits values ('apples', 11)")
	(ODBC:query "insert into fruits values ('oranges', 22)")
	(ODBC:query "insert into fruits values ('bananas', 33)")

	(println "inserted 3 records")

	(ODBC:query "select * from fruits")

	(println "performed a query")

	(println (ODBC:num-cols) " columns in result set")
	(println "fetching rows ...")
	(while (set 'row (ODBC:fetch-row)) 
		(set 'row (map trim row))
		(println row))
	(println)


	(ODBC:query "delete from fruits")
	(println "rows deleted: " (ODBC:affected-rows))
	
	(println "\nclosing database")
	(ODBC:close-db)
	)



;; eof
