;; sqlite3.lsp - module for SQLite3 newLISP bindings
;;
;; SQLite version 3.0 introduced a new database format and is incompatible
;; whith the previous 2.1 to 2.8 format Old SQLite 2.x based databases can
;; be converted  using the old and new sqlite client application:
;;
;;    sqlite OLD.DB .dump | sqlite3 NEW.DB
;;
;; While in sqlite 2.8 all returned fields where of string type, sqlite3
;; retturns, text, integer or float. Blobs are returned as text and NULLs
;; are returned as nil.
;;
;; See http://sqlite.org for details
;;
;; version 1.0 - initial release sqlite3 version
;;               tested on WinXP/freeBSD and LINUX
;;
;; version 1.1 - 'get-integer' to 'get-int'
;;
;; version 1.2 - doc error for sql3:sql, automatic loading of libsqlite3 depending
;;               on current newLISP platform Linux/BSD/MacOSX/Solaris/MinGW-Win32
;;
;;             - sql type FLOAT did not retrieve correctly
;;
;; version 1.3 - now handles BLOB data type
;;
;; version 1.4 - sql3:error set to nil instead "" when no error
;;               failing query now returns nul instead of error message
;;
;; version 1.5 - replaced int for integer
;;
;; required: sqlite3.dll or libsqlite3.so from http://www.sqlite.org
;;
;;
;; USAGE:
;;
;; (sql3:open "MYDB")			; opens/creates a database returns a handle (ignore)
;;					; or 'nil' on failure
;;
;; (sql3:sql "select * from mytable;")	; makes a sql query, returns result
;;
;; (sql3:changes)			; returns number of rows affected
;;
;; (sql3:rowid)				; returns last row id from insert
;;
;; (sql3:tables)			; returns a list of all tables in a database
;;
;; (sql3:columns tableName)		; returns a list of column lables for tableName
;;
;; (sql3:close)				; closes the database
;;
;; (sql3:error)                         ; returns error text
;;
;; (sql3:timeout 10000)			; sets bust timeout to 10,000 ms, returns db handle (ignore)

(context 'sql3)

;; set libray to path-name of the library on your platform OS
;;
(if 
	;; LINUX and BSDs
	(< (& 0xF (last (sys-info))) 3)	(set 'library "/usr/local/lib/libsqlite3.so") 
	;; Mac OSX / Darwin
	(= (& 0xF (last (sys-info))) 3) (set 'library "/usr/lib/libsqlite3.0.dylib")
	;; Solaris
	(= (& 0xF (last (sys-info))) 4)	(set 'library "/usr/local/lib/libsqlite3.so") 
	;; MinGW, Win32
	(> (& 0xF (last (sys-info))) 4) (set 'library "/Program Files/sqlite3/sqlite3.dll")	     
	true (println "Cannot load library OS not supported"))

(import library "sqlite3_open" "cdecl")
(import library "sqlite3_close" "cdecl")
(import library "sqlite3_prepare" "cdecl")
(import library "sqlite3_step" "cdecl")
(import library "sqlite3_column_count" "cdecl")
(import library "sqlite3_column_name" "cdecl")
(import library "sqlite3_column_type" "cdecl")
(import library "sqlite3_column_int" "cdecl")
(import library "sqlite3_column_double" "cdecl")
(import library "sqlite3_column_text" "cdecl")
(import library "sqlite3_column_blob" "cdecl")
(import library "sqlite3_column_bytes" "cdecl")
(import library "sqlite3_finalize" "cdecl")
(import library "sqlite3_get_table" "cdecl")
(import library "sqlite3_last_insert_rowid" "cdecl")
(import library "sqlite3_changes" "cdecl")
(import library "sqlite3_busy_timeout" "cdecl")
(import library "sqlite3_errmsg" "cdecl")


;; gloablly used vars and constants

(define db nil)                  ; database handle
(define dbp "\000\000\000\000")  ; ptr to database handle
(define error-message nil)        ; error message
(define col-names '())           ; list of column headers
(define col-types '())           ; list of column types
(define pstm "\000\000\000\000") ; ptr to compiled sql

(constant 'SQLITE_OK 0)
(constant 'SQLITE_ROW 100)
(constant 'SQLITE_DONE 101)

(constant 'SQLITE_TYPES '(
	0 
	SQLITE_INTEGER 
	SQLITE_FLOAT 
	SQLITE_TEXT 
	SQLITE_BLOB 
	SQLITE_NULL))

;; open or ceate a database ;;;
;; if the database does exist it gets opened, else a new database
;; with the name given is created.
;;
;; if a database already has been opened set error messge
;; and return nil. A subsequernt (sql3:error) will return
;; the error message.
;;
;;
(define (sql3:open db-name)
  ; only open if not alrady done
  (if (not db)
    (begin
      (set 'result (sqlite3_open db-name dbp))
      (if (!= result SQLITE_OK)
        (set 'db nil)
        (set 'db (get-int dbp))))
    (begin
      (set 'error-message "A database is already open")
      nil))
)

;; close the currently open database ;;;
;;
;;
(define (sql3:close) 		;;  overwrite the close in MAIN
	(if db (begin
		(sqlite3_close db)
		(set 'db nil)
		true)))

;; execute an SQL statement ;;;
;;
;; For 'select' statements an array of the results is returned or '()
;; for the empty set. For other statements 'true is returned for a
;; successful outcome. On failure 'nil is returened (sql3:error)
;; can be used to retrieve the error text.
;;
;;
(define (sql sql-str)
	(set 'result nil 'done nil 'error-message nil)
	(set 'sqarray '());
	(set 'col-names '());
	(set 'col-types '());

	;; set up parameters for sqlite3_prepare() call
	(set 'ppstm "\000\000\000\000") ; pointer to statement ptr
	(set 'pptail "\000\000\000\000") ; pointer to statement tail

	;; compile the sql statment
	(if db (set 'result (sqlite3_prepare db sql-str -1 ppstm pptail)))

	;; set up parameters for sqlite3_step() call
	(set 'pstm (get-int ppstm))

	;; execute the compiled statement
	(if (= result SQLITE_OK) 
		(while (not done) 
			;; execute statement until done/101 or 
			(set 'result (sqlite3_step pstm))
			(set 'num-cols (sqlite3_column_count pstm))
			(if (empty? col-names) (set 'col-names (get-names pstm num-cols)))
			(if (empty? col-types) (set 'col-types (get-types pstm num-cols)))
			(if (= result SQLITE_ROW)
				(push (get-values pstm num-cols) sqarray -1)
				(set 'done true) ;; received done/101 or error
			))
		
                )
	;; if done/101 finalize
	(if (= result SQLITE_DONE) 
		(begin
			(set 'result (sqlite3_finalize pstm))
			;; for 'select' statements return the array else 'true'
			(if (> num-cols 0) sqarray true))
		(if (= result 0) true (set-error))))


(define (get-values pstm cols) 
	(set 'row '())
	(dotimes (idx cols)
		(set 'i (int idx)) ; all loop vars are float
		(case (nth idx col-types)
			(SQLITE_INTEGER 
				(push (sqlite3_column_int pstm i) row -1))
			(SQLITE_FLOAT 
				(set 'pstr (sqlite3_column_text pstm i))
				(if (= pstr 0)
					(push nil row -1)
					(push (float (get-string pstr)) row -1)))
			(SQLITE_TEXT 
				(set 'pstr (sqlite3_column_text pstm i))
				(if (= pstr 0)
					(push "" row -1)
					(push (get-string pstr) row -1)))

			;; not tested yet
			(SQLITE_BLOB 
				(set 'pstr (sqlite3_column_blob pstm i))
				(set 'len (sqlite3_column_bytes pstm i))
				(set 'buff (dup "\000" len))
				(if (= pstr 0)
					(push "" row -1)
					(begin
						(cpymem pstr buff len)
						(push buff row -1))))
			(SQLITE_NULL 
				(push nil row -1))))
	row)


(define (get-names pstm cols) 
	(set 'row '())
	(dotimes (idx cols)
		(set 'i (int idx)) ; all loop vars are float
		(set 'ps (sqlite3_column_name pstm i))
		(if (= ps 0)	       ;; check for null pointer to result
			(push "" row -1)
			(push (get-string ps) row -1)))
	row)

(define (get-types pstm cols)
	(set 'row '())
	(dotimes (idx cols)
		(set 'i (int idx)) ; all loop vars are float
		(push (nth (sqlite3_column_type pstm i) SQLITE_TYPES) row -1))
	row)



;; return last row id from last 'insert'
;;
(define (rowid)
	(if db (sqlite3_last_insert_rowid db)))

;; return a list of tables in database
;;
(define (tables)
	(if db (begin
		(set 'lst (sql "select tbl_name from sqlite_master")) ))
		(if lst (set 'lst (first (transpose lst)))))



;; return a list of column names for a table
;;
(define (columns aTable)
        (if (list? (sql (append "select * from " aTable " where 0;")))
                col-names))


;; return number of rows changed/affected
;;
(define (changes)
	(if db (sqlite3_changes db)))



;; sets busy timeout in milliseconds
;;
(define (timeout ms)
	(if db (sqlite3_busy_timeout db ms)))



;; return the error text of the last error occured in sql3:sql
;;
;;
(define (error) error-message)
	
(define (set-error)
	(set 'result (sqlite3_errmsg db))
	(if (= result 0) 
		(set 'error-message nil)
		(set 'error-message (get-string result))
		nil
		)
)


(context 'MAIN)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; test the database routines
;;
;; if there is an old "TEST" db from an earlier sqlite 2.8 delete it first
;;
(define (test-sqlite3)
	(if (sql3:open "TEST") 
		(println "database opened/created,  ... Ok")
		(println "problem opening/creating database"))

	(if (sql3:sql "create table fruits (name CHAR(20), qty INT(3), price FLOAT(10), blobtext BLOB);")
		(println "created table fruits,  ... Ok")
		(println "problem creating table fruits"))

	(if (sql3:sql "insert into fruits values ('apples', 11, 1.234, X'41424300010101');")
		(println "inserted, last row id: " (sql3:rowid) ",  ... Ok")
		(println "problem inserting row"))

	(if (sql3:sql "insert into fruits values ('oranges', 22, 2.345, X'42434400020202');")
		(println "inserted, last row id: " (sql3:rowid) ",  ... Ok")
		(println "problem inserting row"))

	(if (sql3:sql "insert into fruits values ('bananas', 33, 3.456, X'44454600030303');")
		(println "inserted, last row id: " (sql3:rowid) ",  ... Ok")
		(println "problem inserting row"))

	(set 'sqarray (sql3:sql "select * from fruits;"))

	(if sqarray
		(begin
			(println "selected rows: ") 
			(map println sqarray)
			(println "column types: ")
			(map println sql3:col-types)
			(println "... Ok")
			)
		(println "problem with select"))

;[text]
	(if (sql3:sql "delete from fruits where 1;")
		(println "deleted, rows affected: " (sql3:changes) ",  ... Ok")
		(println "problem deleting rows"))

	(if (list? (sql3:tables) )
		(println "tables: " (sql3:tables) ", ... Ok")
		(println "problem in sql3:tables"))

	(if (list? (sql3:columns "fruits") )
		(println "columns: " (sql3:columns "fruits") ", ... Ok")
		(println "problem in sql3:columns"))
	

	(if (sql3:sql "drop table fruits;")
		(println "table fruits dropped,  ... Ok")
		(println "problem dropping table fruits"))


	(sql3:close)
;[/text]
)


;; eof ;;
